import os
import json
import tempfile

from fabric.api import *
from fabric.operations import _prefix_commands, _prefix_env_vars


__all__ = ['sshagent_run', 'call_python', 'ec2_hostnames', 'ec2_instances', 'answer_sudo']


def sshagent_run(cmd, user=None):
    """
    Helper function.
    Runs a command with SSH agent forwarding enabled.

    Note:: Fabric (and paramiko) can't forward your SSH agent.
    This helper uses your system's ssh to do so.
    """
    # Handle context manager modifications
    wrapped_cmd = _prefix_commands(_prefix_env_vars(cmd), 'remote')
    if user is None:
        user = env.user
    opts = ['-A', '-o ServerAliveInterval=60']
    if env.disable_known_hosts:
        opts += ['-o StrictHostKeyChecking=no',
                 '-o UserKnownHostsFile=/dev/null']
    try:
        host, port = env.host_string.split(':')
        opts.append('-p %s' % port)
    except ValueError:
        host = env.host_string
    opts = ' '.join(opts)
    return local("ssh %s %s@%s '%s'" % (opts, user, host, wrapped_cmd))


def call_python(method, *args):
    """
    Call the given ``method'' with the given ``args'' in Python on the
    remote server.  ``method'' should be the full Python path to the
    method.  Only JSON-serializable arguments and return values are
    supported.
    """
    module = '.'.join(method.split('.')[:-1])
    args = json.dumps(args)[1:-1]
    output = run('/usr/bin/env python -c \'import json, {module};'
                 'print json.dumps({method}({args}))\''
                 ''.format(module=module, method=method, args=args))
    return json.loads(output)


def ec2_hostnames(*args, **kwargs):
    """
    Returns a list of hostnames for the specified filters.
    """
    from fabulaws.ec2 import EC2Service
    return EC2Service().public_dns(*args, **kwargs)


def ec2_instances(*args, **kwargs):
    """
    Returns a list of instances for the specified filters.
    """
    from fabulaws.ec2 import EC2Service
    return EC2Service().instances(*args, **kwargs)


def answer_sudo(cmd, *args, **kwargs):
    """
    Answers questions presented via standard out according to the
    given answers.  The ``answers`` keyword argument should be a list of
    (question, answer) pairs.  The questions are in regular expression format,
    so ensure that any special characters are appropriately escaped.
    """
    answers = kwargs.pop('answers', [])
    if answers:
        # use shared memory rather than potentially writing passwords to the
        # most likely unencrypted disk
        tmpdir = os.path.exists('/dev/shm') and '/dev/shm' or None
        script = tempfile.NamedTemporaryFile(dir=tmpdir)
        script.writelines([
            "import pexpect, sys\n",
            "child = pexpect.spawn('{0}')\n".format(cmd),
            # the following line may be enabled for debugging purposes, but
            # will spill passphrases in plain text in log files
            #"child.logfile = sys.stdout\n"
        ])
        for question, answer in answers:
            script.writelines([
                "child.expect('{0}')\n".format(question),
                "child.sendline('{0}')\n".format(answer),
            ])
        # this is important, otherwise pexect will kill the process before it's
        # finished
        script.writelines(["child.wait()\n"])
        script.flush()
        remote_file = '/dev/shm/{0}'.format(os.path.basename(script.name))
        put(script.name, remote_file, mirror_local_mode=True)
        cmd = 'python {0}'.format(remote_file)
    result = sudo(cmd, *args, **kwargs)
    if answers:
        run('rm {0}'.format(remote_file))
    return result



import time

from fabric.api import *
from fabric.contrib import files

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin


class RabbitMqMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that configures RabbitMQ
    """

    package_name = 'rabbitmq'
    rabbitmq_packages = ['rabbitmq-server']
    rabbitmq_ulimit = 20000

    @uses_fabric
    def rabbitmq_service(self, cmd):
        return sudo('service rabbitmq-server {0}'.format(cmd))

    def bind_app_directories(self, *args, **kwargs):
        tries = kwargs.pop('rabbitmq_tries', 10)
        sleep = kwargs.pop('rabbitmq_sleep', 2)
        # make sure we stop before proceeding in case we get moved to a secure directory
        self.rabbitmq_service('stop') 
        super(RabbitMqMixin, self).bind_app_directories(*args, **kwargs)
        # try starting a number of times with warn_only=True, as rabbitmq
        # fails to restart occassionally for unknown reasons
        restarted = False
        for i in range(tries-1):
            with settings(warn_only=True):
                result = self.rabbitmq_service('start')
                if result.return_code == 0:
                    restarted = True
                    break
                else:
                    time.sleep(sleep)
        # if we haven't succeeded yet, start again without warn_only=True
        # and let the failure be handled as usual by Fabric
        if not restarted:
            self.rabbitmq_service('start') 

    @uses_fabric
    def rabbitmq_command(self, command):
        """Run a rabbitmqctl command."""

        sudo(u'rabbitmqctl %s' % command)

    def create_mq_user(self, username, password):
        """Create a rabbitmq user."""

        self.rabbitmq_command(u'add_user %s %s' % (username, password))

    def create_mq_vhost(self, name):
        """Create a rabbitmq vhost."""

        self.rabbitmq_command(u'add_vhost %s' % name)

    def set_mq_vhost_permissions(self, vhost, username, permissions='".*" ".*" ".*"'):
        """Set permssions for a user on a given vhost."""

        self.rabbitmq_command(u'set_permissions -p %s %s %s' % (vhost, username, permissions))

    @uses_fabric
    def rabbitmq_configure(self):
        files.append('/etc/default/rabbitmq-server',
                     'ulimit -n %s' % self.rabbitmq_ulimit, use_sudo=True)
        self.rabbitmq_service('restart')

    def setup(self):
        """Redis mixin"""

        super(RabbitMqMixin, self).setup()
        self.rabbitmq_configure()


class RabbitMqOfficialMixin(RabbitMqMixin):
    """
    Installs RabbitMQ from the official upstream APT repository.
    """

    rabbitmq_aptrepo = ('http://www.rabbitmq.com/debian/', 'testing', 'main', '056E8E56')
    rabbitmq_packages = ['rabbitmq-server']



import datetime

from fabric.api import *
from fabric.contrib import files

from fabulaws.decorators import *
from fabulaws.api import *
from fabulaws.ubuntu.packages.base import AptMixin

class MemcachedMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures memcached.
    """
    package_name = 'memcached'
    memcached_packages = ['memcached']
    memcached_ulimit = 10000
    memcached_bind = '127.0.0.1' # set to '' to bind to all interfaces
    memcached_memory = '64'
    memcached_connections = '1024'
    memcached_conf = '/etc/memcached.conf'
    memcached_init_default = '/etc/default/memcached'
    memcached_conf_patterns = {
        'bind': ('-l', r'\S+'),
        'connections': ('-c', r'\w+'),
        'memory': ('-m', r'\w+'),
        'threads': ('-t', r'\w+')
    }

    @uses_fabric
    def memcached_service(self, cmd):
        sudo('/etc/init.d/memcached {0}'.format(cmd), pty=False)

    @uses_fabric
    def memcached_configure(self, **kwargs):
        for key, (opt, pat) in self.memcached_conf_patterns.iteritems():
            if key in kwargs:
                val = kwargs[key]
            else:
                val = getattr(self, 'memcached_{0}'.format(key), None)
            if val is None:
                continue
            full_pat = '^#?\s*{0}\s+{1}'.format(opt, pat)
            full_val = '{0} {1}'.format(opt, val)
            # we don't know if the config file contains the line or not, so
            # comment it out if it does and then add a new line with the date
            files.comment(self.memcached_conf, full_pat, use_sudo=True)
            if val != '':
                date = datetime.datetime.now().strftime('%Y-%m-%d')
                message = '\n# {0}; added by fabulaws {1}\n{2}'\
                          ''.format(key, date, full_val)
                files.append(self.memcached_conf, message, use_sudo=True)
        if self.memcached_ulimit is not None:
            ulimit = 'ulimit -n {0}'.format(self.memcached_ulimit)
            files.append(self.memcached_init_default, ulimit, use_sudo=True)
        self.memcached_service('restart')

    def setup(self):
        """memcached mixin"""

        super(MemcachedMixin, self).setup()
        self.memcached_configure()



import copy
import random
import urllib2

from fabric.api import *
from fabric.contrib import files

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin


class PythonMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs Python
    """

    package_name = 'python'
    python_ppa = None
    python_packages = ['python', 'python-dev']
    python_install_tools = True
    python_pip_version = None # install the latest version
    python_virtualenv_version = None # install the latest version
    python_pypi_mirrors = [
        'https://pypi.python.org',
    ]

    def _find_mirror(self):
        """
        Finds a PyPI mirror that appears to be online.
        """

        for mirror in self.python_pypi_mirrors:
            try:
                r = urllib2.urlopen(mirror)
                if r.code == 200:
                    return mirror
            except urllib2.URLError:
                pass
        raise ValueError('No active mirror found.')

    @uses_fabric
    def install_python_tools(self):
        """
        Installs the required Python tools from a random (hopefully online)
        PyPI mirror.
        """

        mirror = self._find_mirror()
        version = ''
        if self.python_pip_version:
            version = '==%s' % self.python_pip_version
        sudo("easy_install -i %s/simple/ -U pip%s" % (mirror, version))
        version = ''
        if self.python_virtualenv_version:
            version = '==%s' % self.python_virtualenv_version
        sudo("pip install -i %s/simple/ -U virtualenv%s" % (mirror, version))

    def setup(self):
        """
        Hook into the FabulAWS setup() routine to install Python and related
        tools.
        """
        super(PythonMixin, self).setup()
        if self.python_install_tools:
            self.install_packages(['python-setuptools'])
            self.install_python_tools()
